/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

/*
 * Created on Mar 21, 2005
 *
 */
package com.sunlabs.archsim.fleet.components.switchFabric;

import java.util.Iterator;

import com.sunlabs.archsim.base.common.Parameters;
import com.sunlabs.archsim.base.hierarchy.CompositeEntity;
import com.sunlabs.archsim.base.hierarchy.Entity;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.fleet.common.FleetGlobal;
import com.sunlabs.archsim.gasp.Action;
import com.sunlabs.archsim.gasp.ArbiterAction;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.Function;
import com.sunlabs.archsim.gasp.State;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;


/**
 * @author ib27688
 * 
 * A group of components that form a GasP merge.
 * Data is simply copied from input to output.
 * Arbitration is preformed between data-inputs to the branch.
 *
 */
public class Merge extends Component {
	
	public static final String IN= "In";
	public static final String ACKIN= "ackIn";
	public static final String OUT= "Out";
	public static final String ACKOUT= "ackOut";

	protected static final String INTARBITERACTION= "intArbiterAction";
	protected static final String INTOUT= "intOut";
	protected static final String INTACKOUT= "intAckOut";
	protected static final String INTIN= "intIn";
	protected static final String INTACKIN= "intAckIn";
	protected static final String INTMERGEACTION= "intMergeAction";
	protected static final String INTDATASTATE= "intDataState";
	protected static final String INTACKSTATE= "intAckState";
	
	
	public static final String DELIMITER= "_";
	
	/** the number of inputs */
	protected int numInputs;

	/** the inputs for the data */
	protected InputTerminal[] in;
	/* acknowldegmnet that the input has been received */
	protected OutputTerminal[] ackIn;
		
	/** the output for the data */
	protected OutputTerminal out;
	/** the acknowldegment that the output has been received */
	protected InputTerminal ackOut;
	
	/** The arbiter action - to be initialized in a subclass
	 */
	protected ArbiterAction[] arbiterAction;


	// internal inputs, acks, states
	protected OutputTerminal[] intOut;
	protected InputTerminal[] intAckOut;
	
	protected InputTerminal intIn;
	protected OutputTerminal intAckIn;
	
	// internal action
	protected Action intOutAction;
	
	// internal states - between ArbiterActions and intOutAction
	protected State intDataState;
	protected State intAckState;
	
	// hack for XML interface
	public Merge(String name, Parameters params) {
		super(name, null);
		
		int numIn= SwitchFabricDefaults.MERGE_FANIN_DEF;
		
		Delay arbiterDelayParam= null;
		Delay passAADelayParam= null;
		Delay passIODelayParam= null;
		Delay ackAADelayParam= null;
		Delay ackIODelayParam= null;
		Delay passDelayParam= null;
		Delay ackDelayParam= null;
		
		// process the parameters
		if (params != null) {
			for (int i= 0; i<params.size(); i++) {
				String pName= params.getName(i);
				String sValue= params.getValue(i);
				if (pName.equals(SwitchFabricDefaults.MERGE_FANIN)
						|| pName.equals(FleetGlobal.PARAMETER) ) {
					// was fanin provided properly 
					try {
						numIn= Integer.parseInt(sValue);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " + SwitchFabricDefaults.MERGE_FANIN
								+ ": " + sValue
								+ ". Value is IGNORED and the default value of "
								+ numIn + " is used.");
					}					
				} // if merge fanin
				else if (pName.equals(SwitchFabricDefaults.MERGE_ARBITER_DELAY) ) {
					try {
						int ad= Integer.parseInt(sValue);
						arbiterDelayParam= new Delay(ad);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ SwitchFabricDefaults.MERGE_ARBITER_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED.");
					}					
				}
				else if (pName.equals(SwitchFabricDefaults.MERGE_ARBITER_ACTION_PASS_DELAY) ) {
					try {
						int pd= Integer.parseInt(sValue);
						passAADelayParam= new Delay(pd);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ SwitchFabricDefaults.MERGE_ARBITER_ACTION_PASS_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED.");
					}					
					
				}
				else if (pName.equals(SwitchFabricDefaults.MERGE_INTOUT_ACTION_PASS_DELAY) ) {
					try {
						int pd= Integer.parseInt(sValue);
						passIODelayParam= new Delay(pd);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ SwitchFabricDefaults.MERGE_INTOUT_ACTION_PASS_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED.");
					}					
					
				}
				else if (pName.equals(SwitchFabricDefaults.MERGE_ARBITER_ACTION_FIRED_DELAY) ) {
					try {
						int fd= Integer.parseInt(sValue);
						ackAADelayParam= new Delay(fd);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ SwitchFabricDefaults.MERGE_ARBITER_ACTION_FIRED_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED.");
					}					
					
				}				
				else if (pName.equals(SwitchFabricDefaults.MERGE_INTOUT_ACTION_FIRED_DELAY) ) {
					try {
						int fd= Integer.parseInt(sValue);
						ackIODelayParam= new Delay(fd);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ SwitchFabricDefaults.MERGE_INTOUT_ACTION_FIRED_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED.");
					}					
					
				}				
				else if (pName.equals(SwitchFabricDefaults.MERGE_PASS_DELAY) ) {
					try {
						int pd= Integer.parseInt(sValue);
						passDelayParam= new Delay(pd);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ SwitchFabricDefaults.MERGE_PASS_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED.");
					}					
					
				}	
				else if (pName.equals(SwitchFabricDefaults.MERGE_ACK_DELAY) ) {
					try {
						int ackd= Integer.parseInt(sValue);
						ackDelayParam= new Delay(ackd);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ SwitchFabricDefaults.MERGE_ACK_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED.");
					}					
					
				}	
				else {
					director.error("Component: " + name + ", id= " + getID() 
							+ ": Attribute " + pName + " with value " + sValue
							+ " is ignored");					
				}

			} // for
		} // if parames != null
		
		buildMerge(numIn);
		
		// apply the parammeters
		// non-null's indicate that a value was passed
		if ( arbiterDelayParam != null) {
			setArbitrationDelay(arbiterDelayParam);
		}
		if (passAADelayParam != null) {
			for (int i=0; i< numInputs; i++) {
				((VanillaArbiterAction)arbiterAction[i]).setCopyDelay(passAADelayParam);
			}			
		}
		if ( passIODelayParam != null ) {
			((IntMergeAction)intOutAction).setCopyDelay(passIODelayParam);			
		}
		if (ackAADelayParam != null) {
			for (int i=0; i< numInputs; i++) {
				((VanillaArbiterAction)arbiterAction[i]).setAckDelay(ackAADelayParam);
			}			
		}		
		if (ackIODelayParam != null) {
			((IntMergeAction)intOutAction).setAckDelay(ackIODelayParam);			
		}
		if (passDelayParam != null) {
			setCopyDelay(passDelayParam);
		}
		if (ackDelayParam != null) {
			setAckDelay(ackDelayParam);
		}
		
		// set the internal ack state active
		intAckState.setInitiallyActive(null);
				
	} // Merge(name, parameters)
	
	public Merge(String name) {
		this(name, 2);
	}
	
	
	/** create a Merge - give a number of branches */
	public Merge(String name, int numIn) {
		super(name, null);
		
		// build the merge
		buildMerge(numIn);
		
		// set the internal ack state active
		intAckState.setInitiallyActive(null);
		
	} // Merge constructor

	
	// used in constructors
	protected void buildMerge(int numIn) {
		// set the number of branches, must be at least 1
		numInputs= (numIn > 0)? numIn : 1;

		// make the output terminal and its ack-sibling
		out= new OutputTerminal(OUT);
		addMember(out);
		
		ackOut= new InputTerminal(ACKOUT);
		addMember(ackOut);
		
		// make the terminal arrays - input side first
		in= new InputTerminal[numInputs];
		ackIn= new OutputTerminal[numInputs];
		
		// make internal terminal arrays
		intOut= new OutputTerminal[numInputs];
		intAckOut= new InputTerminal[numInputs];
		
		// make the internal terminals that attach to the
		// internal action on the output side
		intIn= new InputTerminal(INTIN);
		addMember(intIn);
		intAckIn= new OutputTerminal(INTACKIN);
		addMember(intAckIn);
		
		// make the array for the ArbiterAction
		arbiterAction= new ArbiterAction[numInputs];
		
		// make the action on the output side
		intOutAction= new IntMergeAction(INTMERGEACTION);
		addMember(intOutAction);
		
		// attach internal terminals to the internal action on
		// the output side
		intOutAction.attach(intIn);
		intOutAction.attach(intAckIn);
		
		// attach the external terminals on the output side
		intOutAction.attach(out);
		intOutAction.attach(ackOut);
		
		// make the two internal states
		intDataState= new State(INTDATASTATE);
		addMember(intDataState);
		intAckState= new State(INTACKSTATE);
		addMember(intAckState);
		
		
		// attach internal terminals to the internal state -
		// still on the outpout side
		intDataState.attach(intIn);
		intAckState.attach(intAckIn);
		
		// now make ArbiterActions, its associated terminals, 
		// and attach the terminals to the internal states
		for (int i= 0; i< numInputs; i++) {
			// make a pair of terminals for the input side
			in[i]= new InputTerminal(IN + DELIMITER + i, this);
			ackIn[i]= new OutputTerminal(ACKIN + DELIMITER + i, this);
			
			// make the arbter action
			arbiterAction[i]= new VanillaArbiterAction(
					INTARBITERACTION + DELIMITER + i, i, this);
			
			// make a pair of terminals for the output side
			intOut[i]= new OutputTerminal(INTOUT + DELIMITER + i, this);
			intAckOut[i]= new InputTerminal(INTACKOUT + DELIMITER + i, this);
			
			// attach the terminals
			arbiterAction[i].attach(in[i]);
			arbiterAction[i].attach(ackIn[i]);
			arbiterAction[i].attach(intOut[i]);
			arbiterAction[i].attach(intAckOut[i]);
			
			// set the "ack" for the arbiter
			arbiterAction[i].setDoneInput(intAckOut[i]);
			
			// attach the "output side terminals" 
			// to the internal states
			intDataState.attach(intOut[i]);
			intAckState.attach(intAckOut[i]);
			
			// join the arbiter actions in a group
			if (i > 0) {
				arbiterAction[0].attachPeer(arbiterAction[i]);
			} 			
		} // for		
	}
	
	public void setArbitrationDelay(Delay d) {
		for (int i= 0; i<numInputs; i++) {
			arbiterAction[i].setArbiterDelay(d);
		}
	} // setArbitrationDelay
	
	public void setCopyDelay(Delay d) {
		for (int i=0; i< numInputs; i++) {
			((VanillaArbiterAction)arbiterAction[i]).setCopyDelay(d);
		}
		((IntMergeAction)intOutAction).setCopyDelay(d);
	} // setCopyDelay
	
	public void setAckDelay(Delay d) {
		for (int i=0; i< numInputs; i++) {
			((VanillaArbiterAction)arbiterAction[i]).setAckDelay(d);
		}
		((IntMergeAction)intOutAction).setAckDelay(d);
	} // setAckDelay
	
	public InputTerminal getIn(int index) {
		if (index >=0 && index < numInputs) {
			return in[index];
		}
		else {
			return null;
		}
	}
	
	public OutputTerminal getAckIn(int index) {
		if (index >=0 && index < numInputs) {
			return ackIn[index];
		}
		else {
			return null;
		}		
	}
	
	public OutputTerminal getOut() {
		return out;
	}
	
	public InputTerminal getAckOut() {
		return ackOut;
	}

	public OutputTerminal getIntOut(int index) {
		if (index >=0 && index < numInputs) {
			return intOut[index];
		}
		else {
			return null;
		}		
	} // getIntOut
	
	public InputTerminal getIntAckOut(int index) {
		if (index >=0 && index < numInputs) {
			return intAckOut[index];
		}
		else {
			return null;
		}		
	} // getIntAckOut
	
	public InputTerminal getIntIn() {
		return intIn;
	} // getIntIn
	
	public OutputTerminal getIntAckIn() {
		return intAckIn;
	} // getIntAckIn
	
	public ArbiterAction getIntAction(int index) {
		if (index >=0 && index < numInputs) {
			return arbiterAction[index];
		}
		else {
			return null;
		}			
	} // getIntArbAction
	
	public Action getIntOutAction() {
		return intOutAction;
	} // getIntOutAction
	
	public State getIntDataState() {
		return intDataState;
	} // getIntDataState

	public State getIntAckState() {
		return intAckState;
	} // getIntDataState
	
	public Terminal getTerminal(String name) {
		Terminal t=null;
		for (Entity e: members) {
			if (e instanceof Terminal) {
				Terminal tFound= (Terminal)e;
				if (tFound.getBaseName().equals(name)
						|| tFound.getName().equals(name)
						|| tFound.getBaseAlias().equals(name)
						|| tFound.getAlias().equals(name)) {
					t= tFound;
					break;
				}
			}
		}
		return t;
	}
	
} // Merge


class VanillaArbiterAction extends ArbiterAction {

	protected int index;
	
	protected Function copy;
	protected Function acknowledge;
	
	protected String inTerminalName;
	
	public VanillaArbiterAction(String name, int ind, CompositeEntity parent) {
		super(name, parent);
		// store the index
		index=ind; 
		// instantiate the functions
		copy= new PassData();
		acknowledge= new ActionFired();
		
		inTerminalName= Merge.IN+Merge.DELIMITER+index;
		
		// attach functions to terminal names
		outputFunctions.put(Merge.INTOUT+Merge.DELIMITER+index,
							copy);
		outputFunctions.put(Merge.ACKIN+Merge.DELIMITER+index, 
				acknowledge);

		// do we have a global merge arbiter delay set?
		Integer aDelay= globals.intValue(SwitchFabricDefaults.MERGE_ARBITER_DELAY);
		Delay arbDelay= (aDelay != null)
			? new Delay(aDelay.intValue())
			: SwitchFabricDefaults.MERGE_ARBITER_DELAY_DEF;
			
		arbiter.setDelay(arbDelay);
		
	} // VanillaArbiterAction
	
	public void setCopyDelay(Delay d) {
		copy.setDelay(d);
	}
	
	public void setAckDelay(Delay d) {
		acknowledge.setDelay(d);
	}
	
	/** Simply pass the data from input terminal 0 */
	private class PassData extends Function {
		
		public PassData() {
			super("PassData");
			Integer fDelay= globals.intValue(SwitchFabricDefaults.MERGE_PASS_DELAY);
			functionDelay= (fDelay != null)
				? new Delay(fDelay.intValue())
				: SwitchFabricDefaults.MERGE_PASS_DELAY_DEF;
		}
		
		public Object value() {
			Iterator ti= terminals.iterator();
			boolean found= false;
			Terminal t= null;
			while (!found && ti.hasNext()) {
				t= (Terminal)(ti.next());
				if (t.getBaseName().equals(inTerminalName)) {
					found= true;
				}
			}
			if (found) {
				return t.getData();
			}
			else {
				director.fatalError("function " + getName()
						+ " could not found terminal "
						+ inTerminalName
						+ " attached to action "
						+ VanillaArbiterAction.this.getAlias() + ", id= "
						+ VanillaArbiterAction.this.getID());
				director.stop();			
				return null;
			}
		} // value
	} // PassData

	/** Vacuous function used to acknowldge that the action has fired */ 
	private class ActionFired extends Function {
		
		public ActionFired() {
			super("Vacous");
			Integer fDelay= globals.intValue(SwitchFabricDefaults.MERGE_ACK_DELAY);
			functionDelay= (fDelay != null)
				? new Delay(fDelay.intValue())
				: SwitchFabricDefaults.MERGE_ACK_DELAY_DEF;		
		} // constructor
	
		public Object value() {
			return null;
		} // value
	} // AcknowledgeData
	
} // class AATAction


class IntMergeAction extends Action {

	Function outF;
	
	Function ackOutF;
	
	public IntMergeAction(String name) {
		super(name);
		outF= new PassData();
		ackOutF= new ActionFired();
		
		outputFunctions.put(Merge.OUT, outF);
		outputFunctions.put(Merge.INTACKIN, ackOutF);
	} // FifoAction
	
	public void setCopyDelay(Delay d) {
		outF.setDelay(d);
	}
	
	public void setAckDelay(Delay d) {
		ackOutF.setDelay(d);
	}
	
	/** Simply pass the data from input terminal 0 */
	private class PassData extends Function {
		
		public PassData() {
			super("PassData");
			Integer fDelay= globals.intValue(SwitchFabricDefaults.MERGE_PASS_DELAY);
			functionDelay= (fDelay != null)
				? new Delay(fDelay.intValue())
				: SwitchFabricDefaults.MERGE_PASS_DELAY_DEF;
		} // constructor
		
		public Object value() {
			Iterator ti= terminals.iterator();
			boolean found= false;
			Terminal t= null;
			while (!found && ti.hasNext()) {
				t= (Terminal)(ti.next());
				if (t.getBaseName().equals(Merge.INTIN)) {
					found= true;
				}
			}
			if (found) {
				return t.getData();
			}
			else {
				return null;
			}
		} // value
	} // PassData

	/** Vacuous function used to acknowldge that the action has fired */ 
	private class ActionFired extends Function {
		
		public ActionFired() {
			super("Vacous");
			Integer fDelay= globals.intValue(SwitchFabricDefaults.MERGE_ACK_DELAY);
			functionDelay= (fDelay != null)
				? new Delay(fDelay.intValue())
				: SwitchFabricDefaults.MERGE_ACK_DELAY_DEF;		
		}
	
		public Object value() {
			return null;
		} // value
	} // AcknowledgeData
	
} // IntMergeAction
