/**
 * Copyright 2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 */

package com.sunlabs.archsim.fleet.ships.fifo;

/*
 * Created on Feb 2, 2006
 *
 */

import com.sunlabs.archsim.base.common.Parameters;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.FleetGlobal;
import com.sunlabs.archsim.fleet.common.OutputFunction;
import com.sunlabs.archsim.fleet.common.Payload;
import com.sunlabs.archsim.gasp.Action;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.Function;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;


/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

/**
 * This is a basic builidng block for a FifoShip. A string of two FifoShipElement-s
 * forms a one-place fifo, three fifo elements form a two place fifo, ...
 * The difference between a Fifo and a FifoShip is that a Fifo simply passes data
 * through. A FifoShip expects Payload on the input and passes through Data.
 * 
 * Interface: FleetGlobal.IN, FleetGlobal.ACK_IN on the data input side, 
 * and FleetGlobal.OUT, FleetGlobal.ACK_OUT on the output side.
 * 
 * @author ib27688
 */
public class FifoShipElement extends Component {

	InputTerminal in;
	OutputTerminal ackIn;
	OutputTerminal out;
	InputTerminal ackOut;
	
	FifoAction action;
	
	public FifoShipElement(String name, Parameters params) {
		super(name, null);
		
		// process the parameters
		Delay fwdDelayParam= null;
		Delay ackDelayParam= null;
		
		if (params != null) {
			for (int i= 0; i<params.size(); i++) {
				String pName= params.getName(i);
				String sValue= params.getValue(i);
				
				if (pName.equals(FifoDefaults.FIFO_FWD_DELAY) 
						|| pName.equals(FifoDefaults.FWD_DELAY)) {
					try {
						int d = Integer.parseInt(sValue);
						fwdDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ FifoDefaults.FIFO_FWD_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED");						
					}
				}
				else if (pName.equals(FifoDefaults.FIFO_ACK_DELAY) 
						|| pName.equals(FifoDefaults.ACK_DELAY)) {
					try {
						int d = Integer.parseInt(sValue);
						ackDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ FifoDefaults.FIFO_ACK_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED");						
					}
				}
				else {
					director.error("Component: " + name + ", id= " + getID() 
							+ ": Attribute " + pName + " with value " + sValue
							+ " is ignored");					
				} // else
			} // for
		} // if params != null				
		
		buildElement();
		
		// set delays if needed
		if (fwdDelayParam != null) {
			action.passDataFn.setDelay(fwdDelayParam);
		}
		if (ackDelayParam != null) {
			action.ackFn.setDelay(fwdDelayParam);
		}				
	}
	
	public FifoShipElement() {
		// no need for an variable name here
		super("FifoShipElement", null);		
		buildElement();
	} // FifoShipElement
	
	protected void buildElement() {
		action= new FifoAction("FifoAction");
		addMember(action);
		
		in= new InputTerminal(FleetGlobal.IN);
		addMember(in);
		action.attach(in);
		
		ackIn= new OutputTerminal(FleetGlobal.ACK_IN);
		addMember(ackIn);
		action.attach(ackIn);
		
		out= new OutputTerminal(FleetGlobal.OUT);
		addMember(out);
		action.attach(out);
		
		ackOut= new InputTerminal(FleetGlobal.ACK_OUT);
		addMember(ackOut);
		action.attach(ackOut);

	}
	
	public Terminal getTerminal(String name) {
		if (name.equals(FleetGlobal.IN)) {
			return in;
		}
		else if (name.equals(FleetGlobal.ACK_IN)) {
			return ackIn;
		}
		else if (name.equals(FleetGlobal.OUT)) {
			return out;
		}
		else if (name.equals(FleetGlobal.ACK_OUT)) {
			return ackOut;
		}
		else
			return null;
	}

	
	public String stateInfo() {
		
		return "FIFOelement " + getName() + ", id= " + getID() + ",  has passed " 
		+ action.counter
		+ " items.";
	} // stateInfo
	
	class FifoAction extends Action {
		
		int counter= 0;
		
		public PassData passDataFn;
		public ActionFired ackFn;

		public FifoAction(String name) {
			super(name);
			passDataFn= new PassData();
			ackFn= new ActionFired();
			outputFunctions.put(FleetGlobal.OUT, passDataFn);
			outputFunctions.put(FleetGlobal.ACK_IN, ackFn);
		} // FifoAction
		
		
		/** Simply pass the data from input terminal 0 */
		class PassData extends OutputFunction {
			
			public PassData() {
				super("PassData");
				Integer fDelay= globals.intValue(FifoDefaults.FIFO_SHIP_FWD_DELAY);
				functionDelay= (fDelay != null)
					? new Delay(fDelay.intValue())
					: FifoDefaults.FIFO_SHIP_FWD_DELAY_DEF;
			}
			
			public Data value() {
				// increase the count of inputs
				counter++;
				// get data from the input
				Object d=  in.getData();
				// is this a payload object?
				if (d instanceof Payload) {
					// get the data item out
					return ((Payload)d).data;
				}
				else if (d instanceof Data) {
					return (Data)d;
				}
				else {
					// wrong input data type, output an illegal value
					director.error("FifoShipElement " + getAlias() + ", id= " + getID() 
							+ "received data of unexpected data type: " 
							+ d + " of type " + d.getClass().getName()
							);						
					return new Data(d, false);
				}
					
			} // value
		} // PassData

		/** Vacuous function used to acknowldge that the action has fired */ 
		class ActionFired extends Function {
			
			public ActionFired() {
				super("Vacous");
				Integer fDelay= globals.intValue(FifoDefaults.FIFO_SHIP_ACK_DELAY);
				functionDelay= (fDelay != null)
					? new Delay(fDelay.intValue())
					: FifoDefaults.FIFO_SHIP_ACK_DELAY_DEF;
			}
		
			public Data value() {
				return null;
			} // value
		} // AcknowledgeData
		
	} // FifoAction
	
} // FifoShipElement



