// Copyright 2006 all rights reserved; see LICENSE file for BSD-style license

package edu.berkeley.sbp;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import java.lang.ref.*;

/** <font color=green>juxtaposition; zero or more adjacent Elements; can specify a rewriting</font> */
public abstract class Sequence implements Iterable<Element>, SequenceOrElement {

    protected final Element[] elements;

    public boolean needed_or_hated = false;

    final HashSet<Sequence> hated  = new HashSet<Sequence>();

    final HashSet<Sequence> needs  = new HashSet<Sequence>();
    final HashSet<Sequence> hates  = new HashSet<Sequence>();

    final Position          firstp;

    Atom follow = null;

    // Static Constructors //////////////////////////////////////////////////////////////////////////////

    abstract Sequence _clone();
    Sequence dup() {
        Sequence ret = _clone();
        for(Sequence s : needs) { ret.needs.add(s); }
        for(Sequence s : hates) { ret.hates.add(s); s.hated.add(ret); }
        ret.follow = follow;
        return ret;
    }

    /** create an empty sequence (matches the empty string) */
    public static Sequence create() { return new Sequence.Constant.Empty(); }

    /** create a sequence of one element */
    public static Sequence create(Element e) { return create(new Element[] { e }, 0); }

    /** create a sequence which drops the result of all but one of its element */
    public static Sequence create(Element[] e, int which) { return new Singleton(e, which); }

    /** create a sequence which always evaluates to a constant result  */
    public static Sequence create(Element[] e, Object result) { return new Constant(e, result); }

    /**
     *  create a sequence (general form)
     *  @param head   the head of the output tree
     *  @param e      the elements to match
     *  @param drop   only elements of <tt>e</tt> whose corresponding <tt>boolean</tt> in <tt>drops</tt>
     *                is <i>false</i> will be included in the output tree
     *  @param foster if true, all children of the last child (ie
     *                grandchildren) are promoted to children of this
     *                node; this is very useful for matching repetitions
     **/
    public static Sequence create(Object head, Element[] e, boolean[] drop, boolean foster) {
        return foster
            ? new Unwrap(e, head, drop)
            : new RewritingSequence(head, e, drop);
    }

    ////////////////////////////////////////////////////////////////////////////////

    /** return a new sequence identical to this one, but with a positive conjunct <tt>s</tt> */
    public Sequence and(Sequence s) { Sequence ret = dup(); ret.needs.add(s); s.needed_or_hated=true; return ret; }

    /** return a new sequence identical to this one, but with a negative conjunct <tt>s</tt> */
    public Sequence andnot(Sequence s) { Sequence ret = dup(); ret.hates.add(s); s.hated.add(ret); s.needed_or_hated=true; return ret; }

    /** return a new sequence identical to this one, but with a follow-set restricted to <tt>a</tt> */
    public Sequence followedBy(Atom a) { Sequence ret = dup(); ret.follow = a; return ret; }

    boolean hatesAny(Iterable<Sequence> it) {
        if (hates.isEmpty()) return false;
        for(Sequence s : it) if (hates.contains(s)) return true;
        return false;
    }

    Iterable<Sequence> needs() { return needs; }
    Iterable<Sequence> hates() { return hates; }

    Position firstp() { return firstp; }

    public Iterator<Element> iterator()    { return new ArrayIterator<Element>(elements); }
    protected Sequence(Element[] elements) {
        this.elements = elements;
        this.firstp = new Position(0);
    }

    // DO NOT MESS WITH THE FOLLOWING LINE!!!
    private Forest.Many epsilonForm = null;
    Forest epsilonForm(Input.Region loc) {
        if (epsilonForm!=null) return epsilonForm;
        epsilonForm = new Forest.Many();
        epsilonForm.merge(firstp().rewrite(loc, false));
        return epsilonForm;
    }

    protected abstract <T> Forest<T> postReduce(Input.Region loc, Forest<T>[] args, Position p);


    // Position //////////////////////////////////////////////////////////////////////////////

    /** the imaginary position before or after an element of a sequence; corresponds to an "LR item" */
    class Position implements IntegerMappable {

        private Forest zero = null;
        public Forest zero(Input.Region reg) {
            if (zero != null) return zero;
            if (pos > 0) throw new Error();
            return zero = rewrite(reg);
        }


                final int      pos;
        private final Position next;
                final Forest[] holder;
        
        private Position(int pos) {
            this.pos      = pos;
            this.next     = pos==elements.length ? null : new Position(pos+1);
            this.holder   = new Forest[elements.length];
        }

        boolean isFirst() { return pos==0; }

        /** the element immediately after this Position, or null if this is the last Position */
        public Element  element() { return pos>=elements.length ? null : elements[pos]; }

        /** the element which produces the sequence to which this Position belongs */
        public Sequence owner() { return Sequence.this; }

        /** the next Position (the Position after <tt>this.element()</tt>) */
        public Position next() { return next; }

        /** true iff this Position is the last one in the sequence */
        public boolean isLast() { return next()==null; }

        // Position /////////////////////////////////////////////////////////////////////////////////

        final <T> Forest<T> rewrite(Input.Region loc) { return rewrite(loc, true); }
        private final <T> Forest<T> rewrite(Input.Region loc, boolean epsilonCheck) {
            if (epsilonCheck && this==firstp()) return epsilonForm(loc);
            for(int i=0; i<pos; i++) if (holder[i]==null) throw new Error("realbad " + i);
            for(int i=pos; i<elements.length; i++) {
                if (holder[i]==null) holder[i] = elements[i].epsilonForm(loc);
                if (holder[i]==null) throw new Error("bad " + i);
            }
            Forest<T> ret = Sequence.this.postReduce(loc, holder, this);
            //for(int k=0; k<pos; k++) holder[k] = null; // to help GC
            return ret;
        }

        public String   toString() {
            StringBuffer ret = new StringBuffer();
            ret.append("<{");
            for(Position p = Sequence.this.firstp(); p != null; p = p.next()) {
                ret.append(' ');
                if (p==this) ret.append(" | ");
                if (p.element()!=null) ret.append(p.element());
                else                   ret.append(' ');
            }
            ret.append("}>");
            return ret.toString();
        }
        private final int idx = master_position_idx++;
        public int toInt() { return idx; }
    }
    private static int master_position_idx = 0;

    // toString //////////////////////////////////////////////////////////////////////////////

    public String toString() { return toString(new StringBuffer(), false).toString(); }
    StringBuffer toString(StringBuffer sb) { return toString(sb, true); }
    StringBuffer toString(StringBuffer sb, boolean spacing) {
        for(int i=0; i<elements.length; i++) {
            sb.append(elements[i].toString());
            sb.append(' ');
        }
        if (follow != null) {
            sb.append("-> ");
            sb.append(follow);
        }
        for(Sequence s : needs) {
            sb.append("& ");
            sb.append(s);
        }
        for(Sequence s : hates) {
            sb.append("&~ ");
            sb.append(s);
        }
        return sb;
    }


    // Specialized Subclasses //////////////////////////////////////////////////////////////////////////////

    static class Constant extends Sequence {
        private final Object result;
        public Constant(Element[] e, Object result) { super(e); this.result = result; }
        Sequence _clone() { return new Constant(elements, result); }
        public <T> Forest<T> postReduce(Input.Region loc, Forest<T>[] args, Position p) {
            return (Forest<T>)Forest.create(loc, result, null, false);
        }
        static class Drop extends Constant {
            Sequence _clone() { return new Drop(elements); }
            public Drop(Element[] e) { super(e, null); }
        }
        static class Empty extends Sequence.Constant.Drop {
            Sequence _clone() { return new Empty(); }
            public Empty() { super(new Element[] { }); } }
    }

    static class Singleton extends Sequence {
        private final int idx;
        public Singleton(Element e) { this(new Element[] { e }, 0); }
        public Singleton(Element[] e, int idx) { super(e); this.idx = idx; }
        public <T> Forest<T> postReduce(Input.Region loc, Forest<T>[] args, Position p) { return args[idx]; }
        Sequence _clone() { return new Singleton(elements,idx); }
    }

    static class Unwrap extends Sequence {
        private boolean[] drops;
        private final Object tag;
        public Unwrap(Element[] e, Object tag)                  { super(e); this.drops = null; this.tag = tag; }
        public Unwrap(Element[] e, Object tag, boolean[] drops) { super(e); this.drops = drops; this.tag = tag; }
        Sequence _clone() { return new Unwrap(elements, tag, drops); }
        public <T> Forest<T> postReduce(Input.Region loc, Forest<T>[] args, Position p) {
            for(int i=0; i<args.length; i++) if (args[i]==null) throw new Error();
            if (drops==null) return Forest.create(loc, (T)tag, args, true);
            int count = 0;
            for(int i=0; i<drops.length; i++) if (!drops[i]) count++;
            Forest<T>[] args2 = new Forest[count];
            int j = 0;
            for(int i=0; i<args.length; i++) if (!drops[i]) args2[j++] = args[i];
            return Forest.create(loc, (T)tag, args2, true);
        }
    }



    static class RegionRewritingSequence extends RewritingSequence {
        private Functor<Input.Region, Object> tagf;
        public RegionRewritingSequence(Functor<Input.Region,Object> tagfunctor, Element[] e, boolean[] drops) {
            super(null, e, drops);
            this.tagf = tagfunctor;
        }
        public <T> Forest<T> postReduce(Input.Region loc, Forest<T>[] args, Position p) {
            this.tag = tagf.invoke(loc);
            Forest<T> ret = super.postReduce(loc, args, p);
            this.tag = null;
            return ret;
        }
    }

    static class RewritingSequence extends Sequence {
        /*private*/public /*final*/ Object tag;
        private final boolean[] drops;
        private int count = 0;
        Sequence _clone() { return new RewritingSequence(tag, elements, drops); }
        public RewritingSequence(Object tag, Element[] e) { this(tag, e, null); }
        public RewritingSequence(Object tag, Element[] e, boolean[] drops) {
            super(e);
            this.tag = tag;
            this.drops = drops == null ? new boolean[e.length] : drops;
            for(int i=0; i<this.drops.length; i++) if (!this.drops[i]) count++;
        }
        public <T> Forest<T> postReduce(Input.Region loc, Forest<T>[] args, Position p) {
            Forest<T>[] args2 = new Forest[count];
            int j = 0;
            for(int i=0; i<args.length; i++) if (!drops[i]) args2[j++] = args[i];
            //System.out.println("reduce \""+tag+"\"");
            return Forest.create(loc, (T)tag, args2, false);
        }
        public StringBuffer toString(StringBuffer sb, boolean spacing) {
            int len = sb.length();
            if (tag != null)
                sb.append("\""+StringUtil.escapify(tag.toString(),"\"\r\n")+"\":: ");
            super.toString(sb, spacing);
            len = sb.length()-len;
            if (spacing) for(int i=0; i<50-len; i++) sb.append(' ');
            return sb;
        }
    }

}
